from pathlib import Path
from pyproject_pip.pypip import (
    get_requirements_packages,
    modify_requirements,
    modify_pyproject_toml,
    name_and_version,
    find_and_sort,
    get_package_info,
)
from pyproject_pip.create import create_project
import subprocess
import sys
import tomlkit
import click

@click.group(invoke_without_command=True)
@click.pass_context
@click.option(
    "-v", "--hatch-env", default=None, help="Specify the Hatch environment to use"
)
def cli(ctx, hatch_env) -> None:
    if ctx.invoked_subcommand is None:
        show_command(hatch_env)


@cli.command("install")
@click.argument("packages", nargs=-1)
@click.option(
    "-r",
    "--requirements",
    type=click.Path(exists=True),
    help="Install packages from the given requirements file",
)
@click.option("-U", "--upgrade", is_flag=True, help="Upgrade the package(s)")
@click.option(
    "-e", "--editable", is_flag=True, help="Install a package in editable mode"
)
@click.option("--hatch-env", default=None, help="Specify the Hatch environment to use")
@click.option(
    "-g",
    "--dependency-group",
    default="dependencies",
    help="Specify the dependency group to use",
)
def install_command(
    packages, requirements, upgrade, editable, hatch_env, dependency_group
) -> None:
    """Install packages and update requirements.txt and pyproject.toml accordingly.

    Args:
        packages (tuple): Packages to install.
        requirements (str, optional): Path to requirements file. If provided, install packages from this file.
        upgrade (bool, optional): Whether to upgrade the packages. Defaults to False.
        editable (bool, optional): Whether to install a package in editable mode. Defaults to False.
        hatch_env (str, optional): The Hatch environment to use. Defaults to "default".
        dependency_group (str, optional): The dependency group to use. Defaults to "dependencies".
    """
    try:
        if requirements:
            if not Path(requirements).exists():
                click.echo(f"Requirements file {requirements} not found. Creating it.")
                Path(requirements).touch()
            packages = get_requirements_packages(requirements)



        for package in packages:
            package_install_cmd = [sys.executable, "-m", "pip", "install"]
            if editable:
                package_install_cmd.append("-e")
            if upgrade:
                package_install_cmd.append("-U")
            package_install_cmd.append(package)
            subprocess.check_call(package_install_cmd)

            package_name, package_version = name_and_version(package, upgrade=upgrade)
            modify_pyproject_toml(
                package,
                package_version,
                action="install",
                hatch_env=hatch_env,
                dependency_group=dependency_group,
            )
            modify_requirements(package_name, package_version, action="install")

    except subprocess.CalledProcessError as e:
        click.echo(f"Error: Failed to install {package}.", err=True)
        click.echo(f"Reason: {e}", err=True)
        sys.exit(e.returncode)


@cli.command("uninstall")
@click.argument("packages", nargs=-1)
@click.option("--hatch-env", default=None, help="Specify the Hatch environment to use")
@click.option(
    "-g",
    "--dependency-group",
    default="dependencies",
    help="Specify the dependency group to use",
)
def uninstall_command(packages, hatch_env, dependency_group) -> None:
    """Uninstall packages and update requirements.txt and pyproject.toml accordingly.

    Args:
        packages (tuple): Packages to uninstall.
        hatch_env (str, optional): The Hatch environment to use. Defaults to "default".
    """
    for package in packages:
        package_name = package.split("==")[0].split("[")[0]  # Handle extras

        try:
            subprocess.check_call(
                [sys.executable, "-m", "pip", "uninstall", package_name, "-y"]
            )
            modify_requirements(package_name, action="uninstall")
            modify_pyproject_toml(
                package_name,
                action="uninstall",
                hatch_env=hatch_env,
                dependency_group=dependency_group,
            )
            click.echo(f"Successfully uninstalled {package_name}")
        except subprocess.CalledProcessError as e:
            click.echo(f"Error: Failed to uninstall {package_name}.", err=True)
            click.echo(f"Reason: {e}", err=True)
            sys.exit(e.returncode)
        except Exception as e:
            click.echo(
                f"Unexpected error occurred while trying to uninstall {package_name}:",
                err=True,
            )
            click.echo(f"{e}", err=True)
            sys.exit(1)


@cli.command("show")
@click.option("--hatch-env", default=None, help="Specify the Hatch environment to use")
def show_command(hatch_env) -> None:
    """Show the dependencies from the pyproject.toml file.

    Args:
        hatch_env (str, optional): The Hatch environment to use. Defaults to "default".
    """
    try:
        with open("pyproject.toml") as f:
            content = f.read()
            pyproject = tomlkit.parse(content)

        # Determine if we are using Hatch or defaulting to project dependencies
        if (
            "tool" in pyproject
            and "hatch" in pyproject["tool"]
            and hatch_env is not None
        ):
            dependencies = (
                pyproject.get("tool", {})
                .get("hatch", {})
                .get("envs", {})
                .get(hatch_env, {})
                .get("dependencies", [])
            )
        else:
            dependencies = pyproject.get("project", {}).get("dependencies", [])

        if dependencies:
            click.echo("Dependencies:")
            for dep in dependencies:
                click.echo(f"  {dep}")
    except FileNotFoundError:
        click.echo("pyproject.toml file not found.")
        sys.exit(1)
    finally:
        sys.exit(0)

@cli.command("find")
@click.argument("package")
@click.option("--limit", default=5, help="Limit the number of results")
@click.option("--sort", default="downloads", help="Sort key to use")
def find_command(package, limit, sort) -> None:
    """Find a package on PyPI and optionally sort the results.

    Args:
        package (str): The package to search for.
        sort (str, optional): Sort key to use. Defaults to "downloads".
    """
    try:
        packages = find_and_sort(package, limit, sort)
        click.echo("Packages found:")
        print(packages)
    except Exception as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)

@cli.command("info")
@click.argument("package")
@click.option("--verbose", "-v", is_flag=True, help="Show verbose output")
def info_command(package, verbose) -> None:
    """Get information about a package from PyPI.

    Args:
        package (str): The package to get information about.
    """
    try:
        package_info = get_package_info(package, verbose)
        click.echo("Package info:")
        click.echo(package_info)
        description = package_info.pop("description", " ")
        if verbose:
            for line in description.split("\n"):
                click.echo(line)
        
    except Exception as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)

@cli.command("create")
@click.argument("project_name")
@click.argument("author")
@click.option("--description", default="", help="Project description")
@click.option("--deps", default=None, help="Dependencies separated by commas")
def create_command(project_name, author, description, deps) -> None:
    """Create a new Python project.

    Args:
        project_name (str): The name of the project.
        author (str): The author of the project.
        description (str, optional): The description of the project. Defaults to "".
        deps (str, optional): Dependencies separated by commas. Defaults to None.
    """
    try:
        if deps:
            deps = deps.split(",")
        create_project(project_name, author, description, deps)
        click.echo(f"Project {project_name} created successfully.")
    except Exception as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)
    
if __name__ == "__main__":
    cli()



import pytest
from pathlib import Path
from unittest.mock import patch, mock_open, call
from pyproject_pip.create import create_project

@pytest.fixture
def mock_cwd(tmp_path):
    with patch('pyproject_pip.create.getcwd', return_value=str(tmp_path)):
        yield tmp_path

def test_create_project(mock_cwd):
    project_name = "test_project"
    author = "Test Author"
    description = "Test Description"
    deps = ["pytest", "numpy"]

    with patch('pyproject_pip.create.Path.mkdir') as mock_mkdir, \
         patch('pyproject_pip.create.Path.write_text') as mock_write_text, \
         patch('pyproject_pip.create.Path.touch') as mock_touch, \
         patch('pyproject_pip.create.create_pyproject_toml', return_value="mock_pyproject_content") as mock_create_pyproject:

        create_project(project_name, author, description, deps)

        # Check if directories were created
        assert mock_mkdir.call_count == 8  # 5 dirs + 1 for tests
        mock_mkdir.assert_has_calls([
            call(exist_ok=True) for _ in range(7)
        ], any_order=True)

        # Check if files were created with correct content
        assert mock_write_text.call_count == 6  # LICENSE, README.md, pyproject.toml, __about__.py
        mock_write_text.assert_has_calls([
            call(''),  # LICENSE
            call(f'# {project_name}\n\n{description}\n\n## Installation\n\n```bash\npip install {project_name}\n```\n'),  # README.md
            call('mock_pyproject_content'),  # pyproject.toml
            call('__version__ = "0.1.0"')  # __about__.py
        ], any_order=True)

        # Check if __init__.py was touched
        assert mock_touch.call_count == 7

        # Check if create_pyproject_toml was called with correct arguments
        mock_create_pyproject.assert_called_once_with(project_name, author, deps)

def test_create_project_with_local_deps(mock_cwd):
    with patch('pyproject_pip.create.Path.mkdir'), \
         patch('pyproject_pip.create.Path.write_text'), \
         patch('pyproject_pip.create.Path.touch'), \
         patch('pyproject_pip.create.create_pyproject_toml') as mock_create_pyproject:

        create_project("local_project", "Local Author", deps="local")
        mock_create_pyproject.assert_called_once_with("local_project", "Local Author", "local")

def test_create_project_no_deps(mock_cwd):
    with patch('pyproject_pip.create.Path.mkdir'), \
         patch('pyproject_pip.create.Path.write_text'), \
         patch('pyproject_pip.create.Path.touch'), \
         patch('pyproject_pip.create.create_pyproject_toml') as mock_create_pyproject:

        create_project("no_deps_project", "No Deps Author")
        mock_create_pyproject.assert_called_once_with("no_deps_project", "No Deps Author", None)

def test_create_project_existing_directory(mock_cwd):
    with patch('pyproject_pip.create.Path.mkdir') as mock_mkdir, \
         patch('pyproject_pip.create.Path.write_text'), \
         patch('pyproject_pip.create.Path.touch'):

        create_project("existing_project", "Existing Author")
        
        # All mkdir calls should have exist_ok=True
        for call in mock_mkdir.call_args_list:
            assert call[1].get('exist_ok', False) is True


"""Synchronizes requirements and hatch pyproject."""

import logging
import re
import subprocess
import sys
from pathlib import Path
from pyproject_pip.create import create_project
import markdown2
import tomlkit
import requests

def clean_text(md_text):
    """Convert Markdown to clean plain text."""
    # Convert Markdown to HTML using markdown2
    html_description = markdown2.markdown(md_text)

    # Function to remove HTML tags
    def strip_html_tags(html):
        clean = re.compile('<.*?>')
        return re.sub(clean, '', html)

    # Strip HTML tags to get plain text
    plain_text_description = strip_html_tags(html_description)

    # Function to remove reStructuredText (reST) directives and roles
    def strip_rst(text):
        # Remove reST directives and roles
        text = re.sub(r'\.\. .*:: .*', '', text)  # remove directives like .. image:: URL
        text = re.sub(r':\w+:.*', '', text)  # remove roles like :target: URL
        text = re.sub(r'\.\. _.*: .*', '', text)  # remove hyperlink targets like .. _name: URL
        return text

    # Clean the plain text description from reST and other special characters
    clean_text = strip_rst(plain_text_description)
    clean_text = re.sub(r'&nbsp;', ' ', clean_text)  # Replace HTML entities
    # clean_text = re.sub(r'\n{2,}', '\n\n', clean_text)  # Replace multiple newlines with a single newline
    clean_text = clean_text.strip()
    # clean_text = clean_text.replace("\n", " ")

    return clean_text

def get_latest_version(package_name):
    """Gets the latest version of the specified package from PyPI.

    Args:
        package_name (str): The name of the package to fetch the latest version for.

    Returns:
        str or None: The latest version of the package, or None if not found or on error.
    """
    try:
        response = requests.get(f"https://pypi.org/pypi/{package_name}/json")
        response.raise_for_status()  # Raises stored HTTPError, if one occurred.
        data = response.json()
        releases = data.get("releases", {})
        if releases:
            filtered_versions = [version for version in releases if all(part.isdigit() for part in version.split("."))]
            sorted_versions = sorted(filtered_versions, key=lambda x: tuple(map(int, x.split("."))), reverse=True)
            return sorted_versions[0] if sorted_versions else None
    except Exception:
        pass
    return None

def search_package(package_name):
    """Search for a package on PyPI and return the description, details, and GitHub URL if available.

    Args:
        package_name (str): The name of the package to search for.
    
    Returns:
        dict: The package information.
    """
    package_info = {}
    try:
        response = requests.get(f"https://pypi.org/pypi/{package_name}/json")
        response.raise_for_status()  # Raises stored HTTPError, if one occurred.
        data = response.json()
        info = data.get("info", {})
        
        package_info["description"] = info.get("summary", "")
        package_info["details"] = info.get("description", "")
        
        # Get GitHub URL if available
        project_urls = info.get("project_urls", {})
        package_info["github_url"] = next(
            (url for _, url in project_urls.items() if "github.com" in url.lower()), None
        )
    except requests.RequestException as e:
        print(f"HTTP error occurred: {e}")
    except Exception as e:
        print(f"An error occurred: {e}")
    
    return package_info


def get_package_names(query_key):
    """Fetch package names from PyPI search results."""
    search_url = f"https://pypi.org/search/?q={query_key}"
    response = requests.get(search_url)
    response.raise_for_status()
    page_content = response.text
    
    # Extract package names from search results
    start_token = '<a class="package-snippet"'
    end_token = '</a>'
    name_token = '<span class="package-snippet__name">'
    
    package_names = []
    start = 0
    while True:
        start = page_content.find(start_token, start)
        if start == -1:
            break
        end = page_content.find(end_token, start)
        snippet = page_content[start:end]
        name_start = snippet.find(name_token)
        if name_start != -1:
            name_start += len(name_token)
            name_end = snippet.find('</span>', name_start)
            package_name = snippet[name_start:name_end]
            package_names.append(package_name)
        start = end
    
    return package_names


def get_package_info(package_name, verbose=False):
    """Retrieve detailed package information from PyPI JSON API."""
    package_url = f"https://pypi.org/pypi/{package_name}/json"
    response = requests.get(package_url)
    response.raise_for_status()
    package_data = response.json()
    
    info = package_data.get("info", {})

    downloads = package_data.get("downloads", {}).get("last_month", 0)
    

    package_info = {
        "name": info.get("name", ""),
        "version": info.get("version", ""),
        "summary": info.get("summary", ""),
        "downloads": downloads
    }
    if verbose:
        package_info["description"] = clean_text(info.get("description", ""))
    project_urls = info.get("project_urls", {})
    try:
        package_info["github_url"] = next(
            (url for _, url in project_urls.items() if "github.com" in url.lower()), None
        )
    except (StopIteration, TypeError, AttributeError):
        package_info["github_url"] = None

    return package_info


def find_and_sort(query_key, limit=5, sort_key="downloads") -> list:
    """Find and sort potential packages by a specified key.

    Args:
        query_key (str): The key to query for.
        sort_key (str): The key to sort by. Defaults to "downloads".
        
    Returns:
        list: List of packages sorted by the specified key.
    """
    try:
        package_names = get_package_names(query_key)
        packages = []
        for package_name in package_names:
            package_info = get_package_info(package_name)
            packages.append(package_info)
        # Sort the packages by the specified key
        sorted_packages = sorted(packages, key=lambda x: x.get(sort_key, 0), reverse=True)
        return sorted_packages[:limit]

    except requests.RequestException as e:
        print(f"HTTP error occurred: {e}")
        return []
    except Exception as e:
        print(f"An error occurred: {e}")
        return []


def modify_requirements(package_name, package_version=None, action="install") -> None:
    """Modify the requirements.txt file to install or uninstall a package.

    Args:
        package_name (str): The name of the package to install or uninstall.
        package_version (str, optional): The version of the package to install. Defaults to None.
        action (str): The action to perform, either 'install' or 'uninstall'.

    Raises:
        FileNotFoundError: If the requirements.txt file does not exist when attempting to read.
    """
    lines = get_requirements_packages(as_set=False)

    # Extract the base package name and optional extras
    base_package_name, *extras = package_name.split("[")
    extras_str = "[" + ",".join(extras) if extras else ""
    package_line = next((line for line in lines if base_package_name == line.split("[")[0].split("==")[0]), None)

    if action == "install":
        if package_version is not  None:
            new_line = f"{base_package_name}{extras_str}=={package_version}"
        else:
            new_line = f"{base_package_name}{extras_str}"

        if package_line:
            # Replace the line with the same base package name
            lines = [line if base_package_name != line.split("[")[0].split("==")[0] else new_line for line in lines]
        else:
            lines.append(new_line)

    elif action == "uninstall":
        # Remove lines with the same base package name
        lines = [line for line in lines if base_package_name != line.split("[")[0].split("==")[0]]

    # Ensure each line ends with a newline character
    lines = [line + "\n" for line in lines]
    with open("requirements.txt", "w") as f:
        f.writelines(lines)


def is_group(line):
    return "[" in line and "]" in line and "\"" not in line[line.index("["):line.index("]")]


def process_dependencies(line, output_lines):
    lines = []
    if "[" in line and ("]" not in line or "," not in line):
        start = line.index("[")
        lines.append(line[:start + 1])
        line = line[start + 1:]

    deps = line.split(",")
    i = 0
    while i < len(deps):
        dep = deps[i]
        if "[" in dep and "]" not in dep:
            while "]" not in dep:
                dep += "," + deps[i+1]
                i += 1
            lines.append(dep + ",")
            i += 1
            continue
        if dep.endswith("]"):
            dep = dep.replace("]", "")
            lines.append(dep)
            lines.append("]")
            i += 1
            continue
        new_dep = dep + "," if dep != deps[-1] else dep
        if new_dep.strip():
            lines.append(new_dep)
        i += 1
    
    for line in lines:
        output_lines.append("  " + line.strip() if not line.strip().startswith("[") else line.strip())

def write_pyproject(data, filename="pyproject.toml"):
    """Write the modified pyproject.toml data back to the file."""
    original_data = Path(filename).read_text()
    try:
        with Path(filename).open("w") as f:
            toml_str = tomlkit.dumps(data)
            inside_dependencies = False
            inside_optional_dependencies = False

            input_lines = toml_str.splitlines()
            output_lines = []
            for input_line in input_lines:
                line = input_line.rstrip()
                if is_group(line):
                    inside_dependencies = False
                    inside_optional_dependencies = "optional-dependencies" in line
                    output_lines.append(line)
                    continue

                if "]" in line and inside_dependencies and "[" not in line:
                    inside_dependencies = False

                if inside_optional_dependencies:
                    process_dependencies(line, output_lines)
        
                if "dependencies" in line and "optional-dependencies" not in line and "extra-dependencies" not in line and not inside_optional_dependencies:
                    inside_dependencies = True
                    inside_optional_dependencies = False
                    output_lines.append(line[:line.index("[") + 1])
                    line = line[line.index("[") + 1:]

                if inside_dependencies and not inside_optional_dependencies:
                    inside_optional_dependencies = False
                    process_dependencies(line, output_lines)

                elif not inside_dependencies and not inside_optional_dependencies:
                    output_lines.append(line)

            for line in output_lines:
                line: str = line.strip()
                line = line.removesuffix(",") if line.endswith("],") else line
                chunks = line.split("==")
                if len(chunks) > 2:
                    line = "==".join(chunks[:2])
                    if not line.endswith("\","):
                        line += "\","
                    

                f.write(line + "\n")
    except Exception as e:
        print(f"An error occurred while writing to {filename}: {e}")
        with Path(filename).open("w") as f:
            f.write(original_data)

def base_name(package_name):
    """Extract the base package name from a package name with optional extras.

    Args:
        package_name (str): The package name with optional extras.

    Returns:
        str: The base package name without extras.
    """
    return package_name.split("[")[0].split("==")[0]

def name_and_version(package_name, upgrade=False):
    if upgrade:
        version = get_latest_version(base_name(package_name))
        return base_name(package_name), version
    if "==" in package_name:
        return package_name.split("==")
    return package_name, None

def modify_dependencies(dependencies, package_version_str, action):
    """Modify the dependencies list for installing or uninstalling a package.

    Args:
        dependencies (list): List of current dependencies.
        package_version_str (str): Package with version string.
        action (str): Action to perform, either 'install' or 'uninstall'.

    Returns:
        list: Modified list of dependencies.
    """
    dependencies = [dep.strip() for dep in dependencies if base_name(dep) != base_name(package_version_str)]
    if action == "install":
        dependencies.append(package_version_str.strip())
    dependencies.sort(key=lambda x: x.lower())  # Sort dependencies alphabetically
    return dependencies

def modify_pyproject_toml(
    package_name,
    package_version="",
    action="install",
    hatch_env=None,
    dependency_group="dependencies",
) -> None:
    """Modify the pyproject.toml file to update dependencies based on action.

    Args:
        package_name (str): The name of the package to install or uninstall.
        package_version (str, optional): The version of the package. Defaults to "".
        action (str): The action to perform, either 'install' or 'uninstall'.
        hatch_env (str, optional): The Hatch environment to use. Defaults to "default".
        dependency_group (str, optional): The group of dependencies to modify. Defaults to "dependencies".
    """
    pyproject_path = Path("pyproject.toml")
    if not pyproject_path.exists():
        action = input("pyproject.toml not found. Do you want to create it? (y/n): ").lower()
        if "y" in action.lower():
            create_project(input("Enter project name: "), input("Enter author name: "), input("Enter project description: "))
        elif "n" in action.lower() and "y" in input("Check parent dirs? (y/n): ").lower():
            for _ in range(3):
                if pyproject_path.exists():
                    break
                pyproject_path = Path("../pyproject.toml")

            if not pyproject_path.exists():
                print("\033pyproject.toml not found in parent directories.\033[0m")
                sys.exit(1)
   

    with open("pyproject.toml") as f:
        content = f.read()
        pyproject = tomlkit.parse(content)

    is_optional = dependency_group != "dependencies"
    is_hatch_env = hatch_env and "tool" in pyproject and "hatch" in pyproject["tool"]
    if hatch_env and not is_hatch_env:
        raise ValueError("Hatch environment specified but hatch tool not found in pyproject.toml.")

    # Prepare the package string with version if provided
    package_version_str = f"{package_name}{('==' + package_version) if package_version else ''}"
    base_project =  pyproject.get("tool", {}).get("hatch", {}).get("envs", {}).get(hatch_env, {}) if is_hatch_env else pyproject.get("project", {})
    optional_base = pyproject.get("project").get("optional-dependencies", {})
    
    if is_optional:
        dependencies = optional_base.get(dependency_group, [])
        optional_base[dependency_group] = modify_dependencies(dependencies, package_version_str, action)

        all_group = optional_base.get("all", [])
        optional_base["all"] = modify_dependencies(all_group, package_version_str, action)
        pyproject["project"]["optional-dependencies"] = optional_base
    else:
        dependencies = base_project.get("dependencies", [])
        base_project["dependencies"] = modify_dependencies(dependencies, package_version_str, action)
    if is_hatch_env:
        pyproject["tool"]["hatch"]["envs"][hatch_env] = base_project
    else:
        pyproject["project"] = base_project

    write_pyproject(pyproject)


def get_pip_freeze():
    """Get the list of installed packages as a set.

    Returns:
        set: Set of installed packages with their versions.
    """
    result = subprocess.run([sys.executable, "-m", "pip", "freeze", "--local"], stdout=subprocess.PIPE, check=False)
    return set(result.stdout.decode().splitlines())


def is_package_in_requirements(package_name) -> bool:
    """Check if a package is listed in the requirements.txt file.

    Args:
        package_name (str): The name of the package.

    Returns:
        bool: True if the package is listed in requirements.txt, False otherwise.
    """
    if not Path("requirements.txt").exists():
        raise FileNotFoundError("requirements.txt file not found.")
    with open("requirements.txt") as f:
        return any(base_name(package_name) == base_name(line) for line in f)

def get_requirements_packages(requirements="requirements.txt", as_set=True):
    """Get the list of packages from the requirements.txt file.

    Returns:
        set: Set of packages listed in the requirements.txt file.
    """
    with open(requirements) as f:
        lines = [line.strip() for line in f if line.strip() and not line.strip().startswith("#")]
    return set(lines) if as_set else lines


def is_package_in_pyproject(package_name, hatch_env=None) -> bool:
    """Check if a package is listed in the pyproject.toml file for a given Hatch environment.

    Args:
        package_name (str): The name of the package.
        hatch_env (str): The Hatch environment to check in. Defaults to "default".

    Returns:
        bool: True if the package is listed in pyproject.toml, False otherwise.
    """
    if not Path("pyproject.toml").exists():
        raise FileNotFoundError("pyproject.toml file not found.")
    with open("pyproject.toml") as f:
        content = f.read()
        pyproject = tomlkit.parse(content)
    is_hatch_env = hatch_env and "tool" in pyproject and "hatch" in pyproject["tool"]
    if hatch_env and not is_hatch_env:
        raise ValueError("Hatch environment specified but hatch tool not found in pyproject.toml.")
    if is_hatch_env:
        dependencies = pyproject["tool"]["hatch"]["envs"][hatch_env]["dependencies"]
    else:
        dependencies = pyproject.get("dependencies", [])
    return any(package_name in dep for dep in dependencies)



import os
import shutil
import sys
from os import getcwd
from pathlib import Path
from typing import Literal


WORKFLOW_UBUNTU ="""name: "Ubuntu"

on:
  push:
    branches: [main]
  pull_request:
    branches: [main]
  workflow_dispatch:

permissions:
  contents: write
  pull-requests: write

jobs:
  ubuntu:
    name: ${{ matrix.os }}
    runs-on: ${{ matrix.os }}
    strategy:
      matrix:
        os: [ubuntu-20.04, ubuntu-latest]

    steps:
      - name: Checkout
        uses: actions/checkout@v3

      - name: Set up Python 3.11
        uses: actions/setup-python@v3
        with:
          python-version: 3.11

      - name: Run install script
        run: |
            python -m pip install --upgrade pip
            python -m pip install hatch

      - name: Cache packages
        uses: actions/cache@v3
        env:
          cache-name: cache-packages
        with:
          path: ~/.local/bin ~/.local/lib .mbodied/envs/mbodied
          key: ${{ runner.os }}-${{ env.cache-name }}-${{ hashFiles('install.bash') }}
          restore-keys: |
            ${{ runner.os }}-${{ env.cache-name }}-

      - name: Check disk usage
        run: df -h

      - name: Clean up before running tests
        run: |
          # Add commands to clean up unnecessary files
          sudo apt-get clean
          sudo rm -rf /usr/share/dotnet /etc/mysql /etc/php /etc/apt/sources.list.d
          # Add more cleanup commands as needed

      - name: Check disk usage after cleanup
        run: df -h

      - name: Run tests
        run: |
          hatch run pip install '.'
          hatch run test"""

WORKFLOW_MAC = """name: "MacOS | Python 3.12|3.11|3.10"

on:
  push:
    branches: [main]
  pull_request:
    branches: [main]
  workflow_dispatch:

permissions:
  contents: read
  pull-requests: write

jobs:
  test:
    name: Python ${{ matrix.python-version }}
    runs-on: macos-latest
    strategy:
      matrix:
        python-version: ["3.12", "3.11", "3.10"]

    steps:
      - name: Checkout code
        uses: actions/checkout@v3

      - name: Set up Python ${{ matrix.python-version }}
        uses: actions/setup-python@v3
        with:
          python-version: ${{ matrix.python-version }}

      - name: Run install script
        run: |
            python -m pip install --upgrade pip
            python -m pip install hatch

      - name: Cache packages
        uses: actions/cache@v3
        env:
          cache-name: cache-packages
        with:
          path: ~/Library/Caches/Homebrew
          key: ${{ runner.os }}-${{ env.cache-name }}-${{ hashFiles('install.bash') }}
          restore-keys: |
            ${{ runner.os }}-${{ env.cache-name }}-
      - name: Run tests
        run: |
          hatch run pip install '.'
          hatch run test"""


def create_project(project_name, author,description="", deps: list[str] | Literal["local"] | None = None, python_version="3.11"):
    # Create project root directory
    root = Path(getcwd()) 
    project_root = root / project_name
    Path(project_root).mkdir(exist_ok=True)
    # Create main directories
    dirs = ['assets', 'docs', 'examples', 'resources', 'tests']
    for dir in dirs:
        Path(root / dir).mkdir(exist_ok=True)


        # Create __about__.py in project directory
    Path(project_root / '__about__.py').touch(exist_ok=True)
       # Create __init__.py in project directory
    add_cli = True
    if not Path(project_root / '__init__.py').exists() and not Path(project_root / 'main.py').exists():
      Path(project_root / '__init__.py').write_text("from .main import cli\n\n__all__ = ['cli']")
      Path(project_root / 'main.py').write_text("from click import command\n\n@command()\ndef cli() -> None:\n    pass\n\nif __name__ == '__main__':\n    cli()")
      add_cli = False
    else:
      Path(project_root / '__init__.py').touch(exist_ok=True)
      
    if not Path(project_root / '__about__.py').exists():
      Path(project_root / '__about__.py').write_text('__version__ = "0.0.1"')

        # Create files in root
    files = [
        ('LICENSE', ''),
        ('README.md', f'# {project_name}\n\n{description}\n\n## Installation\n\n```bash\npip install {project_name}\n```\n'),
        ('pyproject.toml', create_pyproject_toml(project_name, author, deps, python_version=python_version, add_cli=add_cli)),
        ('requirements.txt', 'click' if add_cli else ''),
    ]
    for file, content in files:
      if Path(root / file).exists() and "y" not in input(f"{file} already exists. Overwrite? (y/n): "):

        print(f"{file} already exists. Skipping...") # noqa
        continue
      Path(project_root / file).touch(exist_ok=True)
      Path(file).write_text(content)



 
    Path('tests').mkdir(exist_ok=True)

    # Create workflows directory
    workflows = root / '.github/workflows'
    workflows.mkdir(exist_ok=True, parents=True)
    if Path(workflows / 'macos.yml').exists() or Path(workflows / 'ubuntu.yml').exists():
        should_overwrite = input("Workflows already exist. Overwrite? (y/n): ")
        if should_overwrite.lower() != 'y':
            print("Won't overwrite workflows. Skipping...")
            return
    Path(workflows / 'macos.yml').touch(exist_ok=True)
    Path(workflows / 'ubuntu.yml').touch(exist_ok=True)
    Path(workflows / 'macos.yml').write_text(WORKFLOW_MAC)
    Path(workflows / 'ubuntu.yml').write_text(WORKFLOW_UBUNTU)


def create_pyproject_toml(project_name, author, desc="", deps=None, python_version="3.11", add_cli=True):
    """Create a pyproject.toml file for a Hatch project."""
    authors = ",".join(['{' + f'name="{a}"' + '}' for a in author.split(",")])
    test_docs = "{tests,docs}"
    deps = ",\n     ".join([f'"{dep}"' for dep in deps]) if deps else ""
    version_str = f"py{python_version.replace('.', '')}"
    cli_str = f"{project_name} ={project_name}:cli" if add_cli else ""
    return f'''[build-system]
requires = ["hatchling"]
build-backend = "hatchling.build"

[project]
name = "{project_name}"
dynamic = ["version"]
description = "{desc}"
readme = "README.md"
requires-python = ">={python_version}"
license = "apache-2.0"
keywords = []
authors = [{authors}]
classifiers = [
"Development Status :: 4 - Beta",
"Programming Language :: Python",
"Programming Language :: Python :: {python_version}",
"Programming Language :: Python :: Implementation :: CPython",
"Programming Language :: Python :: Implementation :: PyPy",
]

dependencies = [
    {deps}
]

[project.optional-dependencies]


[project.urls]
Documentation = "https://github.com/{author}/{project_name}#readme"
Issues = "https://github.com/{author}/{project_name}/issues"
Source = "https://github.com/{author}/{project_name}"

[project.scripts]
{cli_str}

[tool.hatch.version]
path = "{project_name}/__about__.py"

[tool.hatch.metadata]
allow-direct-references = true

[tool.hatch.build.targets.wheel.force-include]
"resources" = "{project_name}/resources"

[tool.hatch.envs.default]
python = "{python_version}"
path = ".{project_name}/envs/{project_name}"
dependencies = [
"pytest",
"pytest-mock",
"pytest-asyncio",
]

[tool.hatch.envs.default.env-vars]

[tool.hatch.envs.conda]
type = "conda"
python = "{python_version}"
command = "conda"
conda-forge = false
environment-file = "environment.yml"
prefix = ".venv/"

[tool.hatch.envs.default.scripts]
test = "pytest -vv --ignore third_party {{args:tests}}"
test-cov = "coverage run -m pytest {{args:tests}}"
cov-report = ["- coverage combine", "coverage report"]
cov = ["test-cov", "cov-report"]

[[tool.hatch.envs.all.matrix]]
python = ["3.10", "3.11", "3.12"]

[tool.hatch.envs.types]
dependencies = [
"mypy>=1.0.0"
]
[tool.hatch.envs.types.scripts]
check = "mypy --install-types --non-interactive {{args:{project_name}/ tests}}"

[tool.coverage.run]
source_pkgs = ["{project_name}", "tests"]
branch = true
parallel = true
omit = ["{project_name}/__about__.py"]

[tool.coverage.paths]
{project_name} = ["{project_name}/"]
tests = ["tests"]

[tool.coverage.report]
exclude_lines = ["no cov", "if __name__ == .__main__.:", "if TYPE_CHECKING:"]

[tool.ruff]
line-length = 120
indent-width = 4
target-version = "{version_str}"

[tool.ruff.lint]
extend-unsafe-fixes = ["ALL"]
select = [
"A", "COM812", "C4", "D", "E", "F", "UP", "B", "SIM", "ISC", "N", "ANN", "ASYNC",
"S", "T20", "RET", "SIM", "ARG", "PTH", "ERA", "PD", "I", "PLW",
]
ignore = [
"D100", "D101", "D104", "D106", "ANN101", "ANN102", "ANN003", "UP009", "ANN204",
"B026", "ANN001", "ANN401", "ANN202", "D107", "D102", "D103", "E731", "UP006",
"UP035", "ANN002",
]
fixable = ["ALL"]
unfixable = []

[tool.ruff.format]
docstring-code-format = true
quote-style = "double"
indent-style = "space"
skip-magic-trailing-comma = false
line-ending = "auto"

[tool.ruff.lint.pydocstyle]
convention = "google"

[tool.ruff.lint.per-file-ignores]
"**/{test_docs}/*" = ["ALL"]
"**__init__.py" = ["F401"]
'''



if __name__ == "__main__":
    project_name = input("Enter project name: ")
    author = input("Enter author name: ")
    description = input("Enter project description: ")
    deps = input("Enter dependencies separated by commas: ").split(',')
    create_project(project_name, author, description, deps)



import pytest
from pyproject_pip.pypip import (
    get_latest_version,
    base_name,
    modify_dependencies,
    is_group,
    process_dependencies,
    search_package,
    find_and_sort
)

def test_get_latest_version():
    assert get_latest_version("pytest") is not None
    # Test with a non-existent package
    assert get_latest_version("non_existent_package_12345") is None

def test_base_name():
    assert base_name("package") == "package"
    assert base_name("package[extra]") == "package"
    assert base_name("package==1.0.0") == "package"

def test_modify_dependencies():
    dependencies = ["package1==1.0.0", "package2==2.0.0"]
    
    # Test install action
    result = modify_dependencies(dependencies, "package3==3.0.0", "install")
    assert "package3==3.0.0" in result
    assert len(result) == 3

    # Test uninstall action
    result = modify_dependencies(dependencies, "package1==1.0.0", "uninstall")
    assert "package1==1.0.0" not in result
    assert len(result) == 1

def test_is_group():
    assert is_group("[group]") is True
    assert is_group("not_a_group") is False
    assert is_group('[group with "quotes"]') is False

def test_process_dependencies():
    output_lines = []
    process_dependencies("dep1, dep2, dep3]", output_lines)
    assert output_lines == ['  dep1,', '  dep2,', '  dep3', '  ]']

    output_lines = []
    process_dependencies("dep1, dep2, dep3", output_lines)
    assert output_lines == ['  dep1,', '  dep2,', '  dep3']

    output_lines = []
    process_dependencies("dep1[extra1,extra2], dep2, dep3", output_lines)
    assert output_lines == ['  dep1[extra1,extra2],', '  dep2,', '  dep3']

    output_lines = []
    process_dependencies("dep1[extra1,extra2], dep2[extra3], dep3", output_lines)
    assert output_lines == ['  dep1[extra1,extra2],', '  dep2[extra3', '  ]', '  dep3']

def test_process_dependencies_multiline():
    output_lines = []
    process_dependencies("dep1,\ndep2,\ndep3", output_lines)
    assert output_lines == ['  dep1,', '  dep2,', '  dep3']

def test_process_dependencies_extras_at_end():
    output_lines = []
    process_dependencies("dep1, dep2[extra1,extra2]", output_lines)
    assert output_lines == ['  dep1,', '  dep2[extra1,extra2],']

def test_process_dependencies_remove_extra_quotes():
    output_lines = []
    process_dependencies('"dep1", "dep2"', output_lines)
    assert output_lines == ['  "dep1",', '  "dep2"']

def test_process_dependencies_trailing_comma():
    output_lines = []
    process_dependencies("dep1, dep2,", output_lines)
    assert output_lines == ['  dep1,', '  dep2,']

def test_search_package():

    # Test with a known package
    package_info = search_package("pytest")
    assert len(package_info["description"]) > 0
    assert len(package_info["details"]) > 0
    assert package_info["github_url"] == "https://github.com/pytest-dev/pytest"

    # Test with a non-existent package
    package_info = search_package("non_existent_package_12345")
    assert package_info == {}

def test_find_and_sort():
    # Test with a valid query key
    packages = find_and_sort("pytest")
    assert isinstance(packages, list)
    assert len(packages) > 0


if __name__ == "__main__":
    pytest.main(["-v", __file__])